; CCLC32 - Chaos Construction Light Controller 32
; version 0.1 date 2009-06-23
; Copyright (C) 2005-2009 Stefan Schuermans <stefan@blinkenarea.org>
; Copyleft: GNU public license V2 - http://www.gnu.org/copyleft/gpl.html
; a BlinkenArea project - http://www.blinkenarea.org/

; clock frequency: 12 MHz (external crystal)

; PA0..1: crystal
; PA2: reset
; PB[0-7]: output bus to latches (output, low)
; PD0: RS232 serial input (RX) (input, pull-up disabled)
; PD1: RS232 serial output (TX) (output, high)
; PD2: unused (output, low)
; PD3: latch enable 1st byte (output, low)
; PD4: latch enable 2nd byte (output, low)
; PD5: latch enable 3rd byte (output, low)
; PD6: latch enable 4th byte (output, low)

; serial protocol: 115200, 8N1
;   input:
;     MCUF 32x1-1/16:
;       message: <header> <data>
;       header: <magic> <height> <width> <channels> <maxval>
;       magic: 0x23 0x54 0x26 0x66
;       height: 0x00 0x01
;       width: 0x00 0x20
;       channels: 0x00 0x01
;       maxval: 0x00 0x0F
;       data: <pixel0> ... <pixel31>
;       pixel: 0x00..0x0F



.INCLUDE        "tn2313def.inc"



; constants
.equ    SER_IDLE_TICKS  =       250             ; number of ticks (20ms) serial port must be idle before animation is started



; IO pins
.equ    PORT_BUS        =       PORTB           ; bus to latches
.equ    PORT_LAT_EN     =       PORTD           ; port with latch enables
.equ    BIT_LAT_EN_0    =       3               ;   1st latch enable
.equ    BIT_LAT_EN_1    =       4               ;   2nd latch enable
.equ    BIT_LAT_EN_2    =       5               ;   3rd latch enable
.equ    BIT_LAT_EN_3    =       6               ;   4th latch enable



; general purpose registers
.def    DATA            =       r16
.def    TMP             =       r17
.def    CNT             =       r18
.def    REG0            =       r0

; general purpose registers for interrupt
.def    I_DATA          =       r19
.def    I_TMP           =       r20
.def    I_CNT           =       r21
.def    I_CNT2          =       r22
.def    I_REG0          =       r1
.def    I_REG1          =       r2
; X register reserved for interrupt

; backup registers for interrupt
.def    I_SREG          =       r3

; current PWM value and counter
.def    PWM_VAL         =       r4              ; current PWM value
.def    PWM_CNT         =       r5              ; PWM counter

; receive state
.equ    R_ST_WAIT       =       0               ; waiting for new message
.equ    R_ST_MCUF_HDR   =       1               ; receiving MCUF header
.equ    R_ST_MCUF_DATA  =       2               ; receiving MCUF data
.def    RECV_STATE      =       r6

; receive position
;  - counts received data bytes of current frame
.def    RECV_POS        =       r7

; number of ticks (20ms) serial port is idle
.def    SER_IDLE        =       r8

; current position in animation
.def    ANIM_POS        =       r9

; current delay in animation in ticks (20ms)
.def    ANIM_DELAY      =       r10



.DSEG
.ORG    0x060

; pixel buffers
;  - 4 buffers used for different purposes
;  - pointers to point to current buffer for a purpose
PIXEL_BUF_0:    .BYTE   16
PIXEL_BUF_1:    .BYTE   16
PIXEL_BUF_2:    .BYTE   16
PIXEL_BUF_3:    .BYTE   16

; pixel buffer pointers
;  - pointers to current pixel buffer for a purpose
;  - pointers stored big endian
PIX_PTR_OUT:    .BYTE   2                       ; output in interrupt
PIX_PTR_NEXT:   .BYTE   2                       ; next frame to output
PIX_PTR_RECV:   .BYTE   2                       ; receive buffer
PIX_PTR_ANIM:   .BYTE   2                       ; animation buffer



.MACRO brnex
        breq    EQ
        rjmp    @0
EQ:
.ENDM

.MACRO brshx
        brlo    LO
        rjmp    @0
LO:
.ENDM



.CSEG
.ORG    0x000
        rjmp    ENTRY                           ; RESET
        reti                                    ; INT0
        reti                                    ; INT1
        reti                                    ; TIMER1_CAPT
        rjmp    INT_OUTPUT                      ; TIMER1_COMPA
        reti                                    ; TIMER1_OVF
        reti                                    ; TIMER0_OVF
        reti                                    ; USART0_RX
        reti                                    ; USART0_UDRE
        reti                                    ; USART0_TX
        reti                                    ; ANALOG_COMP
        reti                                    ; PC_INT
        reti                                    ; TIMER1_COMPB
        reti                                    ; TIMER0_COMPA
        reti                                    ; TIMER0_COMPB
        reti                                    ; USI_START
        reti                                    ; USI_OVERFLOW
        reti                                    ; EE_READY
        reti                                    ; WDT



; animation table
;  - contains animation to play when serial port idle
;  - format:
;      <entry> := <duo pixel 0 1> .. <duo pixel 30 31> <delay> <padding>
;      <duo pixel i j> := <pixel i> << 4 | <pixel j>, byte, 0x00..0xFF
;      <pixel i> := brightness value, nibble, 0x0..0xF
;      <delay> := delay in ticks (20ms), byte, 0x00..0xFF
;      <padding> := zero, byte, 0x00
ANIM_TAB:
.db     0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0, 50, 0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F, 50, 0
ANIM_TAB_END:



; code entry point
ENTRY:
; initialize stack pointer
        ldi     TMP,RAMEND
        out     SPL,TMP
; set clock prescaler to 1:1
        ldi     TMP,1<<CLKPCE
        out     CLKPR,TMP
        ldi     TMP,0
        out     CLKPR,TMP
; setup I/O pins
        ldi     TMP,0x00                        ; PB[0-7]: output, low
        out     PORTB,TMP
        ldi     TMP,0xFF
        out     DDRB,TMP
        ldi     TMP,0x02                        ; PB[2-6]: output, low; PB0: input, pull-up off; PB1: output, high
        out     PORTD,TMP
        ldi     TMP,0x7E
        out     DDRD,TMP
; disable analog comparator
        ldi     TMP,0x80
        out     ACSR,TMP
; set up timer 0 to 20ms interval (12MHz / 1024 / 234 = 50.08Hz -> 20ms)
        ldi     TMP,233                         ; count from 0 to 233
        out     OCR0A,TMP
        ldi     TMP,1<<WGM01                    ; clear timer on compare match
        out     TCCR0A,TMP
        ldi     TMP,1<<CS02|1<<CS00             ; prescaler 1:1024
        out     TCCR0B,TMP
; set up timer 1 to 40us interval (12MHz / 1 / 480 = 25kHz -> 40us)
        ldi     TMP,1                           ; count from 0 to 479
        out     OCR1AH,TMP
        ldi     TMP,223
        out     OCR1AL,TMP
        ldi     TMP,0                           ; no output compare
        out     TCCR1A,TMP
        ldi     TMP,1<<WGM12|1<<CS10            ; clear timer on compare match, prescaler 1:1
        out     TCCR1B,TMP
        ldi     TMP,0                           ; no output compare
        out     TCCR1C,TMP
; enable timer 1 compare match interrupt
        ldi     TMP,1<<OCIE1A
        out     TIMSK,TMP
; enable receiver of USART
        ldi     TMP,1<<U2X                      ; double speed
        out     UCSRA,TMP
        ldi     TMP,1<<RXEN                     ; receiver enabled
        out     UCSRB,TMP
        ldi     TMP,1<<UCSZ1|1<<UCSZ0           ; asynchronous, 8N1: 8 data bits, no parity, 1 stop bit
        out     UCSRC,TMP
        ldi     TMP,0                           ; speed = Fosc / (8 * (UBRR + 1))
        out     UBRRH,TMP                       ;       = 12MHz / (8 * (12 + 1))
        ldi     TMP,12                          ;       = 115384bps (115200bps, 0.16% deviation)
        out     UBRRL,TMP
; enable watchdog (64ms)
        wdr
        ldi     TMP,1<<WDCE|1<<WDE
        out     WDTCR,TMP
        ldi     TMP,1<<WDE|1<<WDP1
        out     WDTCR,TMP
        wdr
; jump to main program
        rjmp    MAIN



; output interrupt
;  - called every 40us by timer 1 compare match interrupt
;  - about 300 cycles (about 25us)
; changes: I_SREG, I_DATA, I_TMP, I_CNT, I_CNT2, I_REG0, I_REG1, X
INT_OUTPUT:
        in      I_SREG,SREG                     ; save SREG
; next PWM step (7 cycles)
        inc     PWM_CNT                         ; next step
        mov     I_TMP,PWM_VAL                   ; I_TMP := 2 * PWM_VAL + 1
        add     I_TMP,PWM_VAL
        inc     I_TMP
        cp      PWM_CNT,I_TMP                   ; PWM_CNT < I_TMP ---> nothing to do
        brlo    INT_OUTPUT_END
        clr     PWM_CNT                         ; restart counting steps
; next PWM value (5 cycles)
        inc     PWM_VAL                         ; next PWM value
        ldi     I_TMP,0x0F                      ; PWM_VAL < 0x0F ---> output subframe
        cp      PWM_VAL,I_TMP
        brlo    INT_OUTPUT_GO
        clr     PWM_VAL                         ; restart PWM value
; next PWM cycle (16 cycles)
        lds     I_DATA,PIX_PTR_OUT+0            ; swap output and next buffer
        lds     I_TMP,PIX_PTR_NEXT+0
        sts     PIX_PTR_OUT+0,I_TMP
        sts     PIX_PTR_NEXT+0,I_DATA
        lds     I_DATA,PIX_PTR_OUT+1
        lds     I_TMP,PIX_PTR_NEXT+1
        sts     PIX_PTR_OUT+1,I_TMP
        sts     PIX_PTR_NEXT+1,I_DATA
; calculate and output new b/w subframe (268 cycles)
INT_OUTPUT_GO:
; get base address of current output buffer
        lds     XH,PIX_PTR_OUT+0
        lds     XL,PIX_PTR_OUT+1
; process 32 pixels (264 cycles)
        ldi     I_CNT2,1<<BIT_LAT_EN_0          ; use bitmask for latch bit as counter
INT_OUTPUT_LOOP_32:
; process 8 pixels (56 cycles)
        ldi     I_CNT,4                         ; 4 times 2 pixels
INT_OUTPUT_LOOP_8:
; process 2 pixels (11 cycles)
        ld      I_TMP,X                         ; read byte, get high nibble
        swap    I_TMP
        andi    I_TMP,0x0F
        cp      PWM_VAL,I_TMP                   ; carry = I_TMP > PWM_VAL
        ror     I_DATA                          ; store bit for pixel
        ld      I_TMP,X+                        ; read byte, get low nibble
        andi    I_TMP,0x0F
        cp      PWM_VAL,I_TMP                   ; carry = I_TMP > PWM_VAL
        ror     I_DATA                          ; store bit for pixel
; bottom of 8 pixel loop
        dec     I_CNT
        brne    INT_OUTPUT_LOOP_8
; output data to bus
        out     PORT_BUS,I_DATA                 ; output data on bus
; latch data to pixels
        in      I_DATA,PORT_LAT_EN              ; get port value
        mov     I_TMP,I_DATA                    ; set latch current bit
        or      I_TMP,I_CNT2
        out     PORT_LAT_EN,I_TMP               ; output latch bit
        out     PORT_LAT_EN,I_DATA              ; take back latch bit
; bottom of 32 pixel loop
        lsl     I_CNT2                          ; next bit in mask
        cpi     I_CNT2,(1<<BIT_LAT_EN_3)+1
        brlo    INT_OUTPUT_LOOP_32
; end of interrupt
INT_OUTPUT_END:
        out     SREG,I_SREG                     ; restore SREG
        reti



; serial reception
; changes: Y, Z, DATA, TMP, CNT, REG0
RECV:
; check for reception
        sbis    UCSRA,RXC
        rjmp    RECV_END                        ; nothing received
; check for errors
        sbic    UCSRA,FE
        rjmp    RECV_ERR                        ; frame error
        sbic    UCSRA,DOR
        rjmp    RECV_ERR                        ; data overrun
; read received byte from USART
        in      DATA,UDR
; check state
        ldi     TMP,RECV_TAB_END-RECV_TAB
        cp      RECV_STATE,TMP
        brlo    RECV_CHECKED                    ; state is valid -> use it
        ldi     TMP,R_ST_WAIT                   ; otherwise -> wait for first byte of header
        mov     RECV_STATE,TMP
RECV_CHECKED:
; jump according to state
        ldi     ZH,high(RECV_TAB)
        ldi     ZL,low(RECV_TAB)
        ldi     TMP,0
        add     ZL,RECV_STATE
        adc     ZH,TMP
        ijmp
RECV_TAB:
        rjmp    RECV_WAIT
        rjmp    RECV_MCUF_HDR
        rjmp    RECV_MCUF_DATA
RECV_TAB_END:

; receive error
RECV_ERR:
        in      TMP,UDR                         ; read data from USART to clear error
        ldi     TMP,R_ST_WAIT                   ; not receiving a message, waiting for header
        mov     RECV_STATE,TMP
RECV_END:
        ret

; start waiting for first byte of header
RECV_DO_WAIT:
        ldi     TMP,R_ST_WAIT                   ; not receiving a message, waiting for header
        mov     RECV_STATE,TMP
        ret

; start waiting for first byte of header and check current byte
RECV_DO_WAIT_CHECK:
        ldi     TMP,R_ST_WAIT                   ; not receiving a message, waiting for header
        mov     RECV_STATE,TMP
        rjmp    RECV_WAIT

; waiting for header
RECV_WAIT:
        ldi     ZH,high(2*RECV_MCUF_HDR_DATA)   ; first byte of MCUF header
        ldi     ZL,low(2*RECV_MCUF_HDR_DATA)
        lpm     TMP,Z
        cp      DATA,TMP
        breq    RECV_WAIT_MCUF
        ret                                     ; not a first byte of any known header
; start receiving MCUF header
RECV_WAIT_MCUF:
        ldi     TMP,R_ST_MCUF_HDR               ; receive MCUF header
        mov     RECV_STATE,TMP
        clr     RECV_POS                        ; clear receive position
        rjmp    RECV_MCUF_HDR

; expected MCUF header data (MCUF header 1x5-3/256)
RECV_MCUF_HDR_DATA:
        .db     0x23, 0x54, 0x26, 0x66          ; magic
        .db     0x00, 0x01                      ; height: 1
        .db     0x00, 0x20                      ; width: 32
        .db     0x00, 0x01                      ; channels: 1
        .db     0x00, 0x0F                      ; maxval: 15

; received a MCUF header byte
RECV_MCUF_HDR:
; check header byte
        ldi     ZH,high(2*RECV_MCUF_HDR_DATA)   ; load base address of MCUF header data ("2*" to convert word address to byte address)
        ldi     ZL,low(2*RECV_MCUF_HDR_DATA)
        clr     TMP                             ; add offset
        add     ZL,RECV_POS
        adc     ZH,TMP
        lpm     REG0,Z                          ; fetch expected data byte from program memory
        cp      DATA,REG0                       ; compare with received data byte
        brne    RECV_DO_WAIT_CHECK              ; mismatch -> start waiting for first byte of header and check current byte
; next header byte
        inc     RECV_POS
        ldi     TMP,12                          ; check if entire header has been received
        cp      RECV_POS,TMP
        brsh    RECV_MCUF_HDR_END               ; entire header received -> start receiving data
        ret
; end of header: start receiving MCUF data
RECV_MCUF_HDR_END:
        ldi     TMP,R_ST_MCUF_DATA              ; receive MCUF data
        mov     RECV_STATE,TMP
        clr     RECV_POS                        ; clear receive position
        ret

; received a MCUF data byte
RECV_MCUF_DATA:
; check received value
        cpi     DATA,0x10                       ; values must be smaller 0x10
        brsh    RECV_DO_WAIT_CHECK              ; invalid value-> start waiting for first byte of header and check current byte
; store recevied value in receive buffer
        lds     ZH,PIX_PTR_RECV+0               ; get base address of receive buffer
        lds     ZL,PIX_PTR_RECV+1
        mov     TMP,RECV_POS                    ; get destination address (dest_addr := base_addr + pos / 2)
        lsr     TMP
        add     ZL,TMP
        clr     TMP
        adc     ZH,TMP
        sbrc    RECV_POS,0                      ; jump if position is odd (low nibble)
        rjmp    RECV_MCUF_DATA_ODD
        swap    DATA                            ; even pixel: store received value in upper nibble and clear lower nibble
        st      Z,DATA
        rjmp    RECV_MCUF_DATA_END
RECV_MCUF_DATA_ODD:
        ld      TMP,Z                           ; odd pixel: merge value into lower nibble
        or      TMP,DATA
        st      Z,TMP
RECV_MCUF_DATA_END:
; next channel
        inc     RECV_POS
; check if frame is complete
        ldi     TMP,32                          ; 32 pixels (16 bytes)
        cp      RECV_POS,TMP
        brsh    RECV_MCUF_DATA_COMPLETE         ; end of frame -> use new data for output
        ret

; use receieved data for output
RECV_MCUF_DATA_COMPLETE:
        clr     SER_IDLE                        ; serial port is not idle
; swap receive and next buffer
        cli                                     ; disable interrupts
        lds     DATA,PIX_PTR_RECV+0
        lds     TMP,PIX_PTR_NEXT+0
        sts     PIX_PTR_RECV+0,TMP
        sts     PIX_PTR_NEXT+0,DATA
        lds     DATA,PIX_PTR_RECV+1
        lds     TMP,PIX_PTR_NEXT+1
        sts     PIX_PTR_RECV+1,TMP
        sts     PIX_PTR_NEXT+1,DATA
        sei                                     ; enable interrupts
; start waiting for next header
        rjmp    RECV_DO_WAIT



; tick procedure - called every 20ms
; changes: Y, Z, DATA, TMP, CNT 
TICK:
; check if serial port is already idle long enough to run animation
        ldi     TMP,SER_IDLE_TICKS
        cp      SER_IDLE,TMP
        brsh    TICK_RUN_ANIM                   ; idle long enough -> run animation
; increment idle tick counter of serial port
        inc     SER_IDLE
; check serial port is now idle long enough to run animation
        ldi     TMP,SER_IDLE_TICKS
        cp      SER_IDLE,TMP
        brsh    TICK_START_ANIM                 ; idle long enough -> start animation
; done
        ret

; start animation
TICK_START_ANIM:
        clr     ANIM_POS                        ; set position in animation to begin
        clr     ANIM_DELAY                      ; no delay in animation active
; continue with running animation

; run animation
TICK_RUN_ANIM:
; check if delay is active
        ldi     TMP,0
        cp      ANIM_DELAY,TMP
        breq    TICK_RUN_ANIM_NO_DELAY
; decrement delay
        dec     ANIM_DELAY
; done
        ret

; run animation, no delay active
TICK_RUN_ANIM_NO_DELAY:
; get address of current position in animation table
        mov     ZH,ANIM_POS                     ; Z := ANIM_POS * 18
        swap    ZH
        mov     ZL,ZH
        cbr     ZL,0x0F
        cbr     ZH,0xF0
        clr     TMP
        add     ZL,ANIM_POS
        adc     ZH,TMP
        add     ZL,ANIM_POS
        adc     ZH,TMP
        ldi     TMP,low(2*ANIM_TAB)             ; Z := Z + ANIM_TAB
        add     ZL,TMP
        ldi     TMP,high(2*ANIM_TAB)
        adc     ZH,TMP
; read new values and new delay from animation table
        lds     YH,PIX_PTR_ANIM+0               ; get address of animation buffer
        lds     YL,PIX_PTR_ANIM+1
        ldi     CNT,16                          ; read values for 32 pixels (two pixels per byte)
TICK_RUN_ANIM_NO_DELAY_LOOP:
        lpm     TMP,Z+
        st      Y+,TMP
        dec     CNT
        brne    TICK_RUN_ANIM_NO_DELAY_LOOP
        lpm     ANIM_DELAY,Z+                   ; read delay
; next position in animation
        inc     ANIM_POS
        ldi     TMP,(2*ANIM_TAB_END - 2*ANIM_TAB) / 18  ; number of entries in anim table
        cp      ANIM_POS,TMP
        brlo    TICK_RUN_ANIM_NO_END
        clr     ANIM_POS                        ; set position to begin if end of table reached
TICK_RUN_ANIM_NO_END:
; swap animation and next buffer
        cli                                     ; disable interrupts
        lds     DATA,PIX_PTR_ANIM+0
        lds     TMP,PIX_PTR_NEXT+0
        sts     PIX_PTR_ANIM+0,TMP
        sts     PIX_PTR_NEXT+0,DATA
        lds     DATA,PIX_PTR_ANIM+1
        lds     TMP,PIX_PTR_NEXT+1
        sts     PIX_PTR_ANIM+1,TMP
        sts     PIX_PTR_NEXT+1,DATA
        sei                                     ; enable interrupts
; done
        ret



; main program - first initializations
MAIN:
        wdr                                     ; watchdog reset
        ldi     TMP,0xFF                        ; start new PWM cycle at next interrupt
        mov     PWM_CNT,TMP
        mov     PWM_VAL,TMP
        ldi     TMP,R_ST_WAIT                   ; not receiving a message, waiting for header
        mov     RECV_STATE,TMP
        ldi     TMP,SER_IDLE_TICKS              ; serial port is idle
        mov     SER_IDLE,TMP
        clr     ANIM_POS                        ; set position in animation to begin
        clr     ANIM_DELAY                      ; no delay in animation active

; main program - initialize pixel buffer pointers
        ldi     TMP,high(PIXEL_BUF_0)
        sts     PIX_PTR_OUT+0,TMP
        ldi     TMP,low(PIXEL_BUF_0)
        sts     PIX_PTR_OUT+1,TMP
        ldi     TMP,high(PIXEL_BUF_1)
        sts     PIX_PTR_NEXT+0,TMP
        ldi     TMP,low(PIXEL_BUF_1)
        sts     PIX_PTR_NEXT+1,TMP
        ldi     TMP,high(PIXEL_BUF_2)
        sts     PIX_PTR_RECV+0,TMP
        ldi     TMP,low(PIXEL_BUF_2)
        sts     PIX_PTR_RECV+1,TMP
        ldi     TMP,high(PIXEL_BUF_3)
        sts     PIX_PTR_ANIM+0,TMP
        ldi     TMP,low(PIXEL_BUF_3)
        sts     PIX_PTR_ANIM+1,TMP

; main program - initialize next buffer to zeros
        clr     TMP
        lds     ZH,PIX_PTR_NEXT+0
        lds     ZL,PIX_PTR_NEXT+1
        ldi     CNT,16
MAIN_INIT_OUT_ZERO_LOOP:
        st      Z+,TMP
        dec     CNT
        brne    MAIN_INIT_OUT_ZERO_LOOP

; main program - last initializations
        sei                                     ; enable interrupts



; main program - main loop
MAIN_LOOP:

; watchdog reset
        wdr

; receive serial data
        rcall   RECV

; check for compare match of timer 0
        in      TMP,TIFR
        sbrs    TMP,OCF0A
        rjmp    MAIN_NO_TICK
; call tick procedure
        rcall   TICK
; clear timer 0 compare match flag
        ldi     TMP,1<<OCF0A
        out     TIFR,TMP
MAIN_NO_TICK:

; bottom of main loop
        rjmp    MAIN_LOOP

